#include "amyc_ns.h"
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <regex.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* include for db1 */
#ifdef HAVE_DB1_DB_H
#  include <db1/db.h>
#else
#  ifdef HAVE_DB_185_H
#    include <db_185.h>
#  else
#    include <db.h>
#  endif /* HAVE_DB_185_H */
#endif /* HAVE_DB1_DB_H */

typedef struct cache_entry cache_entry;

static void print_cache_entry (FILE * output, amyc_prefs prefs, cache_entry * entry);
static int key_data_to_cache_entry (cache_entry * entry, DBT * key, DBT * data);
static int cache_entry_cmp (cache_entry * * nn, cache_entry * * pp); 
static char *extract_title (char *filename);

struct cache_entry {
    char * url;
    char * filename;
    char * contenttype;
    char * charset;
    time_t lastmodified;
    time_t filedate;
    time_t expires;
    u_int32_t contentlength;
    u_int32_t reallength;
    int secint;  /* a pointer to a struct containing info about security should be added */
};

int sort_criteria;

int 
remove_url (amyc_prefs prefs, char * URL)
{
  DBT	key;
  DB *db;
  DBTYPE dbtype;
  int fd, i, j, retour, size;
  char filename[500];
  char *strtmp;

    sprintf (filename, "%s/index.db", amyc_cachedirectory (prefs));
    strtmp = give_correct_path ((char *)filename);
/* Check if we can read the cache file */
    fd = open (strtmp, O_RDONLY);
    if (fd == -1)
    {
	 perror (strtmp);
#warning FIXME: we need better error handling
	 return;
    }
    close (fd);

    dbtype = DB_HASH;
    db = dbopen (strtmp, O_RDWR, 0644, dbtype, NULL);
    if (db == NULL) {
	perror ("dbopen");
	exit (EXIT_FAILURE);
    }

    /* Making the key from the URL */
    key . size = strlen (URL) + 1 + 3*4;	/* with '\0' */
    key . data = (char *) malloc((strlen(URL)+1+3*4)*sizeof(char));
    sprintf((key.data),"%c%c%c%c%c%c%c%c%s%c%c%c%c",strlen(URL)+1+3*4,0,0,0,strlen(URL)+1,0,0,0,URL,0,0,0,0);

    retour = (db -> del (db, & key, 0));
    db->close(db);
    return retour;
}

static void 
print_cache_entry (FILE * output, amyc_prefs prefs, cache_entry * entry)
{
    char * tmp_str;
    char filename[255];		/* is 255 enough for a filename ? 
				   or should the len be precisely compputed */
    char *title;
    size_t tmp_str_len;

    /* Print nothing if the "No Image" option is enabled */
    if ((amyc_noimg (prefs)) && !strncmp ("image", entry->contenttype, 5)) {
      return;
    }
	
    fprintf (output, "<TR>");

    if (amyc_showsize (prefs)) {
	fprintf (output, "<TD ALIGN=RIGHT>");
	fprintf (output, "%u\t", entry->reallength);
    }
    if (amyc_showlocal (prefs)) {
	fprintf (output, "<TD>");
	fprintf (output, "<A HREF=\"file:%s/.netscape/cache/%s\">%s</A> \t",
		 getenv ("HOME"),	/* get home directory for correct location of cached files */
		 entry->filename,
		 entry->filename);
    }
    if (amyc_showdate (prefs)) {
	fprintf (output, "<TD>");
	tmp_str_len = 10;
	tmp_str = (char *) malloc (tmp_str_len * sizeof (char));
	strftime (tmp_str, tmp_str_len, "%m/%d/%y", localtime (&(entry->filedate)));
	fprintf (output, "%s   ", tmp_str);
	free (tmp_str);
    }
    if (amyc_showtype (prefs)) {
	fprintf (output, "<TD>");
	if (entry->secint != 0x0201)
	    fprintf (output, "%s\t", entry->contenttype);
    }
    fprintf (output, "<TD>");

    if ((amyc_extitle (prefs)) && !strcmp ("text/html", entry->contenttype)) {
	sprintf (filename, "%s/.netscape/cache/%s",
		 getenv ("HOME"),	/* get home directory for correct location of cached files */
		 entry->filename);
	title = extract_title (filename);
	if (title != NULL) {
	    fprintf (output, "<B>%s</B><BR>\n", title);
	    free (title);	/* Deallocate the string dynamically allocated by extract_title() */
	}
    }
    fprintf (output, "<A HREF=\"%s\">%s</A>\n",
	     entry->url,
	     entry->url);
}

static int 
key_data_to_cache_entry (cache_entry * entry, DBT * key, DBT * data)
{

#ifdef DEBUG
    int i;
#endif
    int headint;
    int decal;
    int offset;

    /* unsigned long is at least 32 bits but probably longer on most architecture, 
       so we use u_int32_t */
    u_int32_t taille;
    u_int32_t tailleURL;
    u_int32_t tailled;
    u_int32_t taillfilename;

    taille = (((unsigned char *) key->data)[0])
	+ (((unsigned char *) key->data)[1] << 8)
	+ (((unsigned char *) key->data)[2] << 16)
	+ (((unsigned char *) key->data)[3] << 24);

#ifdef DEBUG
    fprintf (stdout, "\nKEY: %d\n DATA: %d\n", (int) key->size, (int) data->size);
    fprintf (stdout, "Taille de la cle: %u \n", taille);
#endif
     if (taille ==  1599360585 ) {
       return 1; /* Say that it's a special netscape INT_ entry */
     }
     else {
       /* Si pas entry speciale du type INT_qqch */

	tailleURL = (((unsigned char *) key->data)[4])
	    + (((unsigned char *) key->data)[5] << 8)
	    + (((unsigned char *) key->data)[6] << 16)
	    + (((unsigned char *) key->data)[7] << 24);
#ifdef DEBUG
	fprintf (stdout, "Taille de l'URL: %u \n", tailleURL);
#endif
	entry->url = (char *) malloc (tailleURL * sizeof (char));
	strcpy (entry->url, ((char *) (key->data) + 8));
#ifdef DEBUG
	fprintf (stdout, "URL: %s \n", entry->url);
#endif
	/* Portable (alors que la copie des 4 octets a l'adr de u long int ne le serait pas */
	offset = 0;

	tailled = (((unsigned char *) data->data)[offset++]);
	tailled += (((unsigned char *) data->data)[offset++] << 8);
	tailled += (((unsigned char *) data->data)[offset++] << 16);
	tailled += (((unsigned char *) data->data)[offset++] << 24);

	offset = 8;

	entry->lastmodified = (((unsigned char *) data->data)[offset++]);
	entry->lastmodified += (((unsigned char *) data->data)[offset++] << 8);
	entry->lastmodified += (((unsigned char *) data->data)[offset++] << 16);
	entry->lastmodified += (((unsigned char *) data->data)[offset++] << 24);
#ifdef DEBUG
	fprintf (stdout, "Last modified: %s", ctime (&(entry->lastmodified)));
#endif
	entry->filedate = (((unsigned char *) data->data)[offset++]);
	entry->filedate += (((unsigned char *) data->data)[offset++] << 8);
	entry->filedate += (((unsigned char *) data->data)[offset++] << 16);
	entry->filedate += (((unsigned char *) data->data)[offset++] << 24);
#ifdef DEBUG
	fprintf (stdout, "Date (du fichier): %s", ctime (&(entry->filedate)));
#endif
	entry->expires = (((unsigned char *) data->data)[offset++]);
	entry->expires += (((unsigned char *) data->data)[offset++] << 8);
	entry->expires += (((unsigned char *) data->data)[offset++] << 16);
	entry->expires += (((unsigned char *) data->data)[offset++] << 24);
#ifdef DEBUG
	if (entry->expires)
	    fprintf (stdout, "Expires: %s", ctime (&(entry->expires)));
	else
	    fprintf (stdout, "Expires: No expiration date sent \n");
#endif
	entry->contentlength = (((unsigned char *) data->data)[offset++]);
	entry->contentlength += (((unsigned char *) data->data)[offset++] << 8);
	entry->contentlength += (((unsigned char *) data->data)[offset++] << 16);
	entry->contentlength += (((unsigned char *) data->data)[offset++] << 24);
#ifdef DEBUG
	fprintf (stdout, "Content length: %u \n", entry->contentlength);
#endif
	taillfilename = (((unsigned char *) data->data)[29]);
#ifdef DEBUG
	printf ("Long nom de fichier: %d \n", taillfilename);
#endif
	offset = 33;

	entry->filename = (char *) malloc (tailled * sizeof (char));
	assert (entry->filename);
	strcpy (entry->filename, ((char *) (data->data) + offset));
#ifdef DEBUG
	fprintf (stdout, "Local filename: %s \n", entry->filename);
#endif
	offset += strlen (entry->filename) + 1;

	/* Try to know if there is security for this doc */
	entry->secint = (((unsigned char *) data->data)[offset + 4]);
	entry->secint += (((unsigned char *) data->data)[offset + 5] << 8);
#ifdef DEBUG
	if (entry->secint == 0x0201)
	    printf ("Security: secure (%d) \n", entry->secint);
	else {
	  printf("SecInt: %u", entry->secint);
	  printf("Security: insecure \n"); 
	}
#endif
  
	/* Try to know if there is a special header, when there is the content type is far */
	headint = (((unsigned char *) data->data)[offset+4+9]);
	headint += (((unsigned char *) data->data)[offset+5+9] << 8) ;
	
	if (headint == 0x1) {
	  decal = (((unsigned char *) data->data)[offset+4+9+4]);  
	  decal += (((unsigned char *) data->data)[offset+4+9+5] << 8); 
	  decal += (((unsigned char *) data->data)[offset+4+9+6] << 16); 
	  decal += (((unsigned char *) data->data)[offset+4+9+7] << 24);
#ifdef DEBUG
	  printf("Decal: %u\n", decal);
#endif
	  offset += decal;
	}
#ifdef DEBUG
	printf("HeadInt: %u\n", headint);
#endif
	
	offset += 25;

	/* quel est la taille de la chaine representant le content type ? */
	entry->contenttype = (char *) malloc (30 * sizeof (char));
	strncpy (entry->contenttype, ((char *) (data->data) + (offset)), 30);	/* strncpy -> more security */
#ifdef DEBUG
	fprintf (stdout, "Content type: %s \n", entry->contenttype);
#endif

	offset += strlen (entry->contenttype);

	/* Skip the sub-type (for example gzip when type is application/x-tar */
	if ( (((unsigned char *) data->data)[offset+1]) != 0 ) {
	  offset += ((unsigned char *) data->data)[offset+1];
	}

	if ( (((unsigned char *) data->data)[offset+5]) != 0 ) { /* l'ancienne valeur (11), c'etait la longueur d'un charset particulier le iso-88... */
	    /* extraction du charset */
	    /* quel est la taille de la chaine representant le charset */
	    entry->charset = (char *) malloc (30 * sizeof (char));
	    strncpy (entry->charset, ((char *) (data->data) + (offset + 9)), 30);	/* strNcpy is more secure */
#ifdef DEBUG
	    fprintf (stdout, "Charset: %s \n", entry->charset);
#endif
	    offset += strlen (entry->charset) + 1;	/* Strlen doesn't count the final \0 */
	}
	else {
#ifdef DEBUG
	    fprintf (stdout, "Charset: unknown\n");
#endif
	}

	offset += 10;

	entry->reallength = (((unsigned char *) data->data)[offset++]);
	entry->reallength += (((unsigned char *) data->data)[offset++] << 8);
	entry->reallength += (((unsigned char *) data->data)[offset++] << 16);
	entry->reallength += (((unsigned char *) data->data)[offset++] << 24);

#ifdef DEBUG
	if (entry->reallength != entry->contentlength) {
	    fprintf (stdout, "Real length: %u \n", entry->reallength);
	}
#endif

    }				/* end test to know if it was a special INT_ entry */

#ifdef DEBUG
    for (i = 0; i < key->size; i++) {
	/* affichage caractere par caractere de la clef, car afficher direct avec %s ne marche pas car des \0 apparaissent avant la fin */
	fprintf (stdout, "%c", ((char *) key->data)[i]);
    }
    fprintf (stdout, "\nDATA --> \n");
    fprintf (stdout, "\nTaille extraite: %u \n", tailled);
    for (i = 0; i < data->size; i++) {
	/* affichage caractere par caractere des donnes, car afficher direct avec %s ne marche pas car des \0 appraissent avant la fin */
	fprintf (stdout, "%c", ((char *) data->data)[i]);
    }
#endif

    return 0;
}

void 
amyc_ns_seq (amyc_prefs prefs, FILE *output)
{
    DB *db;
    DBTYPE dbtype;
    DBT key;
    DBT data;
    int fd, i, j, retour, size;
    char filename[500];
    char *strtmp;
    cache_entry **tentry;

    sprintf (filename, "%s/index.db", amyc_cachedirectory (prefs));
    strtmp = give_correct_path ((char *)filename);
/* Check if we can read the cache file */
    fd = open (strtmp, O_RDONLY);
    if (fd == -1)
    {
	 perror (strtmp);
#warning FIXME: we need better error handling
	 return;
    }
    close (fd);

    dbtype = DB_HASH;
    db = dbopen (strtmp, O_RDWR, 0644, dbtype, NULL);
    if (db == NULL) {
	perror ("dbopen");
	exit (EXIT_FAILURE);
    }

    size = 50;
    tentry = (cache_entry * *) malloc (size * sizeof(cache_entry *)); 

    i = 0;

    sort_criteria = amyc_sortby (prefs); /* sort_criteria is the only global var */

    for (retour = db->seq (db, &key, &data, R_FIRST);
	 retour == 0;
	 retour = db->seq (db, &key, &data, R_NEXT)) {
	tentry[i] = (cache_entry *) malloc (sizeof(cache_entry));
	if (!key_data_to_cache_entry(*(tentry+i),&key,&data)) {
#ifndef DEBUG
	    /* print_cache_entry(&prefs,*(entry+i));   */
#endif	
	    i++; 
	    /* Enlarge the tab of pointer on cache_entry by 50 */
	    if (i>=size) {
		size += 50;
		tentry = realloc(tentry,size*sizeof(cache_entry *));
		/* printf("%d",size); */
	    }
	}
    }

    if (sort_criteria != SORT_NOSORT )
	qsort(tentry,i,sizeof(cache_entry *),cache_entry_cmp);  
    
    j = 0;
    for (j = 0; j<i; j++) {
#ifndef DEBUG 
	print_cache_entry(output,prefs,*(tentry+j));   
#endif
    }

    db->close (db);
}

#define INTCMP(a,b)  (a-b)
#define TIMECMP(a,b) INTCMP(a,b) /* The sort criteria "Date" does not always works if "difftime(a,b)" is used instead of "TIMECMP(a,b)" */

static int 
cache_entry_cmp (cache_entry * * nn, cache_entry * * pp)
{
    cache_entry *n, *p;
    n = *nn;
    p = *pp;

    /* use the global var sort_criteria */

    switch (sort_criteria) {
    case SORT_BY_FILEDATE:
        return TIMECMP(n->filedate,p->filedate);
	break;
    case SORT_BY_URL:
	return strcmp (n->url, p->url);
	break;
    case SORT_BY_TYPEURL:
	if (!strcmp (n->contenttype, p->contenttype))
	    return strcmp (n->url, p->url);
	else
	    return strcmp (n->contenttype, p->contenttype);
	break;
    case SORT_BY_TYPEDATE:
	if (!strcmp (n->contenttype, p->contenttype)) {
	  return TIMECMP(n->filedate,p->filedate);
	}
	else
	    return strcmp (n->contenttype, p->contenttype);
	break;
    case SORT_BY_TYPESIZE:
	if (!strcmp (n->contenttype, p->contenttype)) {
	   return INTCMP(n->reallength,p->reallength);
	}
	else
	    return strcmp (n->contenttype, p->contenttype);
	break;
    case SORT_BY_SIZEURL:
	if (n->reallength == p->reallength)
	    return strcmp (n->url, p->url);
	else 
	   return INTCMP(n->reallength,p->reallength);
	break;
    default:
	break;
    }
    return 1;			/* avoid a warning */
}

static char *
extract_title (char *filename)
{
    FILE *f;
    char line[LINE_LEN];
    int erreur;
    regex_t mon_expr;
    char *pattern = "<title>.*</title>";
    int start;
    int stop;
    char *result;

    size_t nmatch = 1;
    regmatch_t matchptr[1];
    erreur = regcomp (&mon_expr, pattern, REG_ICASE | REG_NEWLINE);
    /*  printf("erreur: %d ",erreur); */

    /* Open the HTML file  */
    if ((f = fopen (filename, "r")) == NULL)
	return NULL;

    /* Read the file line by line, search the for the <title> */
    while ((fgets (line, LINE_LEN, f) != NULL) &&
	   (erreur = regexec (&mon_expr, line, nmatch, matchptr, 0))) {
	/* erreur = regexec(&mon_expr, line, nmatch, matchptr, 0); */
    }

    if (!erreur) {
	start = matchptr[0].rm_so + 7;
	stop = matchptr[0].rm_eo - 8;
	result = (char *) malloc ((stop - start + 1) * sizeof (char));
	strncpy (result, line + start, (stop - start));
	result[stop - start] = '\0';

	/*printf("%d %d \n",matchptr[0].rm_so,matchptr[0].rm_eo);
	   printf("%d \n",erreur); */

	return result;
    }
    else
	return NULL;
}
