#include "amyc_prefs.h"
#include "amyc_ns.h"

#include <config.h>

#if ENABLE_NLS
#include <libintl.h>
#  define _(String) gettext (String)
#else
#  define _(String) (String)
#endif /* ENABLE_NLS */

#include <assert.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

/* Browser */

#define BR_NETSCAPE 1
#define BR_OPERA    2

/* For prefs reading */

#define SECTION_HEADER_AMYC "[About My Cache]"
#define SECTION_HEADER_NS   "[Netscape]"
#define SECTION_HEADER_MSIE "[MSIE]"
#define SECTION_HEADER_OP   "[Opera]"

#define PR_RESULT_DIR    "Result Directory"
#define PR_SORT_BY       "Sort by"
#define PR_SHOW_TYPE     "Show type"
#define PR_EXTRACT_TITLE "Extract Title"
#define PR_SHOW_LOCAL    "Show local name"
#define PR_SHOW_DATE     "Show Date"
#define PR_BROWSER       "Browser"
#define PR_SHOW_SIZE     "Show size"
#define PR_NO_IMG        "NoIMG"
#define PR_URL_PAGE      "Url/Page"
#define PR_CACHE_DIR     "Cache Directory"
#define PR_RUN_BROWSER   "Run Browser"

#define PREFS_FILE ".amycrc"

#define LINE_LEN 256

static char * GetPrivateProfileString (char *section, char *field, char *configfilename);
static int GetPrivateProfileInt (char *section, char *field, char *configfilename);
static void print_header (FILE *output, amyc_prefs prefs);
static void print_footer (FILE *output);

struct amyc_prefs {
    int browser;
    int extitle;
    int noimg;
    int showdate;
    int showlocal;
    int showsize;
    int showtype;
    int sortby;
    int urlpage;
    char *cachedirectory;
    char *runbrowser;
    char *resultdirectory;
    char *prefs_file;
};

amyc_prefs
amyc_new (void)
{
    char *strtmp = getenv ("HOME");
    amyc_prefs tmp = malloc (sizeof (struct amyc_prefs));

    tmp->prefs_file = malloc (sizeof (char) * (2 + strlen (strtmp) + sizeof PREFS_FILE));
    strcpy (tmp->prefs_file, strtmp);
    strcat (tmp->prefs_file, "/" PREFS_FILE);

    amyc_get_prefs (tmp);

    return tmp;
}

void
amyc_get_prefs (amyc_prefs prefs)
{
    int used_default_value = 0; 

    prefs->browser = GetPrivateProfileInt (SECTION_HEADER_AMYC, PR_BROWSER, prefs->prefs_file);
    prefs->showdate = GetPrivateProfileInt (SECTION_HEADER_AMYC, PR_SHOW_DATE, prefs->prefs_file);
    prefs->sortby = GetPrivateProfileInt (SECTION_HEADER_AMYC, PR_SORT_BY, prefs->prefs_file);
    prefs->showsize = GetPrivateProfileInt (SECTION_HEADER_AMYC, PR_SHOW_SIZE, prefs->prefs_file);
    prefs->showtype = GetPrivateProfileInt (SECTION_HEADER_AMYC, PR_SHOW_TYPE, prefs->prefs_file);
    prefs->showlocal = GetPrivateProfileInt (SECTION_HEADER_AMYC, PR_SHOW_LOCAL, prefs->prefs_file);
    prefs->noimg = GetPrivateProfileInt (SECTION_HEADER_AMYC, PR_NO_IMG, prefs->prefs_file);
    prefs->extitle = GetPrivateProfileInt (SECTION_HEADER_AMYC, PR_EXTRACT_TITLE, prefs->prefs_file);

    prefs->resultdirectory = GetPrivateProfileString (SECTION_HEADER_AMYC, PR_RESULT_DIR, prefs->prefs_file);

    switch (prefs->browser) {
    case (BR_NETSCAPE):
	prefs->cachedirectory = GetPrivateProfileString (SECTION_HEADER_NS, PR_CACHE_DIR, prefs->prefs_file);
	prefs->runbrowser = GetPrivateProfileString (SECTION_HEADER_NS, PR_RUN_BROWSER, prefs->prefs_file);
	break;
    case (BR_OPERA):
	prefs->cachedirectory = GetPrivateProfileString (SECTION_HEADER_OP, PR_CACHE_DIR, prefs->prefs_file);
	prefs->runbrowser = GetPrivateProfileString (SECTION_HEADER_OP, PR_RUN_BROWSER, prefs->prefs_file);
	break;
    default:
	/* Problem */
	assert (0);
	break;
    }

    /* Some default configuration strings */
    if (prefs->runbrowser == NULL) {
	prefs->runbrowser = malloc (LINE_LEN * sizeof (char));
	sprintf (prefs->runbrowser, "%s", "netscape");
	used_default_value = 1;
    }
    if (prefs->cachedirectory == NULL) {
	prefs->cachedirectory = malloc (LINE_LEN * sizeof (char));
	sprintf (prefs->cachedirectory, "%s", "~/.netscape/cache");
	used_default_value = 1;
    }
    if (prefs->resultdirectory == NULL) {
	prefs->resultdirectory = malloc (LINE_LEN * sizeof (char));
	sprintf (prefs->resultdirectory, "%s", "/tmp");
	used_default_value = 1;
    }

    if (used_default_value == 1)
	amyc_put_prefs (prefs);

#ifdef DEBUG
    printf ("%d \n", prefs->browser);
    printf ("%d \n", prefs->sortby);
    printf ("%d \n", prefs->showtype);
    printf ("%d \n", prefs->showlocal);
    printf ("%d \n", prefs->showdate);
    printf ("%d \n", prefs->noimg);
    printf ("%d \n", prefs->extitle);
    printf ("%s \n", prefs->cachedirectory);
    printf ("%s \n", prefs->runbrowser);
    printf ("%s \n", prefs->resultdirectory);
#endif
}

void
amyc_put_prefs (amyc_prefs prefs)
{
    FILE *f;

    f = fopen (prefs->prefs_file, "w");

    fprintf (f, SECTION_HEADER_AMYC "\n");
    fprintf (f, PR_RESULT_DIR "=%s\n", prefs->resultdirectory);
    fprintf (f, PR_SORT_BY "=%d\n", prefs->sortby);
    fprintf (f, PR_SHOW_TYPE "=%d\n", prefs->showtype);
    fprintf (f, PR_EXTRACT_TITLE "=%d\n", prefs->extitle);
    fprintf (f, PR_SHOW_LOCAL "=%d\n", prefs->showlocal);
    fprintf (f, PR_SHOW_DATE "=%d\n", prefs->showdate);
    fprintf (f, PR_BROWSER "=%d\n", prefs->browser);
    fprintf (f, PR_SHOW_SIZE "=%d\n", prefs->showsize);
    fprintf (f, PR_NO_IMG "=%d\n", prefs->noimg);
    fprintf (f, PR_URL_PAGE "=%d\n", prefs->urlpage);
    fprintf (f, "\n");
    fprintf (f, SECTION_HEADER_NS "\n");
    fprintf (f, PR_CACHE_DIR "=%s\n", prefs->cachedirectory);
    fprintf (f, PR_RUN_BROWSER "=%s\n", prefs->runbrowser);

    fclose (f);
}

/* accesors */

const char *
amyc_cachedirectory (amyc_prefs prefs)
{
    return prefs->cachedirectory;
}

const char *
amyc_resultdirectory (amyc_prefs prefs)
{
    return prefs->resultdirectory;
}

const char *
amyc_getrunbrowser (amyc_prefs prefs)
{
    return prefs->runbrowser;
}

int
amyc_extitle (amyc_prefs prefs)
{
    return prefs->extitle;
}

int
amyc_noimg (amyc_prefs prefs)
{
    return prefs->noimg;
}

int
amyc_showdate (amyc_prefs prefs)
{
    return prefs->showdate;
}

int
amyc_showlocal (amyc_prefs prefs)
{
    return prefs->showlocal;
}

int
amyc_showsize (amyc_prefs prefs)
{
    return prefs->showsize;
}

int
amyc_showtype (amyc_prefs prefs)
{
    return prefs->showtype;
}

int
amyc_sortby (amyc_prefs prefs)
{
    return prefs->sortby;
}

/* Some modifiers */

void set_resultdirectory(amyc_prefs prefs, char * value)
{
  free(prefs->resultdirectory);
  prefs->resultdirectory = (char *) malloc(LINE_LEN*sizeof(char));
  prefs->resultdirectory = strcpy(prefs->resultdirectory, value);
}

void set_cachedirectory(amyc_prefs prefs, char * value)
{
  free(prefs->cachedirectory);
  prefs->cachedirectory = (char *) malloc(LINE_LEN*sizeof(char));
  prefs->cachedirectory = strcpy(prefs->cachedirectory, value);
}

void set_runbrowser(amyc_prefs prefs, char * value)
{
  free(prefs->runbrowser);
  prefs->runbrowser = (char *) malloc(LINE_LEN*sizeof(char));
  prefs->runbrowser = strcpy(prefs->runbrowser, value);
}

void set_showdate(amyc_prefs prefs, int value)
{
  prefs->showdate = value;
}

void set_showsize(amyc_prefs prefs, int value)
{
  prefs->showsize = value;
}

void set_showlocal(amyc_prefs prefs, int value)
{
  prefs->showlocal = value;
}

void set_showtype(amyc_prefs prefs, int value)
{
  prefs->showtype = value;
}

void set_noimg(amyc_prefs prefs, int value)
{
  prefs->noimg = value;
}

void set_extitle(amyc_prefs prefs, int value)
{
  prefs->extitle = value;
}

void set_sortby(amyc_prefs prefs, int value)
{
  prefs->sortby = value;
}

/* End of modifiers */

void
amyc_seq (amyc_prefs prefs, int use_stdout)
{
    FILE *output;
    char filename[500];
    char *strtmp;

    output = stdout; /* default value */
    if (!use_stdout) {
	sprintf (filename, "%s/result.html", prefs->resultdirectory);
	strtmp = give_correct_path ((char *)filename);
	output = fopen (strtmp, "w");
	if (output == NULL) {
	    perror (strtmp);
#warning TOTO: better error handling
	    return;
	}
    }

#ifndef DEBUG
    print_header (output, prefs);
#endif
    amyc_ns_seq (prefs, output);
#ifndef DEBUG
    print_footer (output);
#endif
    if (!use_stdout)
	fclose (output);
}

void 
amyc_runbrowser (amyc_prefs prefs)
{
    char tmp_string[500];

    printf (_("Loading the result in Netscape...\n"));
    sprintf (tmp_string, "%s -remote 'openURL(file:%s/result.html)' \n",
	     give_correct_path (prefs->runbrowser),
	     give_correct_path (prefs->resultdirectory));

    /* Try to open an URL in an already running Netscape process
       If netscape is not running, start a netscape process */
    if (system (tmp_string)) {
	sprintf (tmp_string, "%s %s/result.html &\n",
		 give_correct_path (prefs->runbrowser),
		 give_correct_path (prefs->resultdirectory));
	system (tmp_string);
    }
}

char *
give_correct_path (char *path)
{
    char *return_path, *home_dir, *tmp, *name;
    struct passwd *my_passwd;

    return_path = path;
    if (*path == '~') {
	if (*(path + 1) == '/' || *(path + 1) == '\0') {
	    home_dir = getenv ("HOME");
	    if (home_dir == NULL)
		return return_path;
	}
	else {
	    tmp = strchr (path, '/');
	    if (tmp == NULL)
		my_passwd = getpwnam (path + 1);
	    else {
		name = malloc (sizeof (char) * (tmp - path));
		memcpy (name, path + 1, tmp - path - 1);
		name[tmp - path - 1] = '\0';
		my_passwd = getpwnam (name);
		free (name);
		path = tmp - 1;
	    }
	    if (my_passwd == NULL) {
		return return_path;
	    }
	    home_dir = my_passwd->pw_dir;
	}
	return_path = malloc (sizeof (char) * (strlen (home_dir) + strlen (path)));
	strcpy (return_path, home_dir);
	strcat (return_path, path + 1);
    }

    return return_path;
}

/*
  The section About My Cache is the line strictly equal to "[ABout My Cache]\n".
  If a line starts with "[....]", it is another section.
 */
static char *
GetPrivateProfileString (char *section, char *field, char *configfilename)
{
    int inside_section;
    FILE *f;
    char *line;
    /* char * tmp; */
    char *return_value;
    /*  char * section_header; */

    return_value = NULL;	/* Default value */
    line = (char *) malloc (LINE_LEN * sizeof (char));

    /* Ouvrir configfilename */
    if ((f = fopen (configfilename, "r")) == NULL)
	return NULL;

    inside_section = 0;
    /*  section_header = (char *) malloc ( (strlen(section)+3) * sizeof(char) );
	section_header[0] = '[';
	strcpy((char *) section_header+1,section);
	section_header[strlen(section)+1] = ']';
	printf("%s",section_header); */

    /* Lire le fichier de config ligne par ligne */
    while (fgets (line, LINE_LEN, f) != NULL) {

	if (inside_section && (!strncmp (line, "[", 1))) {
	    inside_section = 0;
	    /* printf("---Fin de section---\n"); */
	}
	if (inside_section) {
	    if (!strncmp (line, field, strlen (field))) {
		return_value = (char *) malloc ( strlen(line) * sizeof (char));
		return_value = strncpy (return_value, line + strlen(field) + 1, 
					strlen(line)-strlen(field)-2 );
		return_value[strlen(line)-strlen(field)-2] = '\0';
		/* fprintf(stdout,"%d \n",atoi(tmp));
		   return_value = atoi(tmp);
		   free(tmp); */
	    }
	}
	if (!strncmp (line, section /*_header*/ , strlen (section /*_header*/ ))) {
	    inside_section = 1;
	    /* printf("---Debut de section---\n"); */
	}
	/* printf("%s",line); */
    }

    fclose(f);
    return return_value;
}

static int 
GetPrivateProfileInt (char *section, char *field, char *configfilename)
{
    int result_value;
    char *rtn_string;
    rtn_string = GetPrivateProfileString (section, field, configfilename);
    if (rtn_string != NULL)
	result_value = atoi (rtn_string);
    else
	return 1; /* default value */
    free (rtn_string);
    return result_value;
}

static void
print_header (FILE *output, amyc_prefs prefs)
{
    fprintf (output, "<HTML>\n");
    fprintf (output, "<HEAD>\n");
    fprintf (output, "<META NAME=\"GENERATOR\" CONTENT=\"" AMYC_NAME " " VERSION "\">\n");
    fprintf (output, "<META NAME=\"AUTHOR\" CONTENT=\"Francois Crevola\">\n");
    fprintf (output, "<META NAME=\"DESCRIPTION\" CONTENT=\"\">\n");
    fprintf (output, "<TITLE>" AMYC_NAME "</TITLE>\n");
    fprintf (output, "<SCRIPT LANGUAGE=\"JAVASCRIPT\">\n");
    fprintf (output, "<!--\n");
    fprintf (output, "function About()\n");
    fprintf (output, " { window.status = \"" AMYC_NAME " %s Unix " VERSION " - Francois Crevola  2001\"; }\n",
	     _("for"));
    fprintf (output, "// -->\n");
    fprintf (output, "</SCRIPT>\n");
    fprintf (output, "</HEAD>\n");
    fprintf (output, "<BODY bgcolor=\"#ffffff\" ONLOAD=\"window.setTimeout('About()',0);\">\n");
    fprintf (output, "<FONT SIZE=\"-1\"><A HREF=\"http://www.crevola.com/about/\">[%s]</A></FONT><BR>\n",
	     _("Official Site"));
    fprintf (output, "<CENTER><FONT SIZE=\"+5\" COLOR=\"RED\">" AMYC_NAME "</FONT><BR>\n");
    fprintf (output, "<FONT SIZE=\"+1\"> 2001 <A HREF=\"mailto:francois@crevola.com?subject=" AMYC_NAME " " VERSION "\">Fran&ccedil;ois Crevola</A></FONT></CENTER>\n");
    fprintf (output, "<H3>%s :</H3>\n",
	     _("These URLs have been found in your cache"));
    fprintf (output, "%s ....<BR><BR>\n",
	     _("If you want"));

    fprintf (output, "<TABLE BORDER=0>\n<TR>");
    if (prefs->showsize)
	fprintf (output, "<TD ALIGN=CENTER BGCOLOR=\"#CCCCCC\">%s", _("Size"));
    if (prefs->showlocal)
	fprintf (output, "<TD ALIGN=CENTER BGCOLOR=\"#CCCCCC\">%s", _("Filename"));
    if (prefs->showdate)
	fprintf (output, "<TD ALIGN=CENTER BGCOLOR=\"#CCCCCC\">%s", _("Date"));
    if (prefs->showtype)
	fprintf (output, "<TD ALIGN=CENTER BGCOLOR=\"#CCCCCC\">%s", _("Type"));
    fprintf (output, "<TD ALIGN=CENTER BGCOLOR=\"#CCCCCC\">URL");
}

static void 
print_footer (FILE * output)
{
    fprintf (output, "</TABLE></BODY></HTML>\n");
}
